from typing import Type
from cleo.commands.command import Command
from poetry.console.application import Application
from poetry.plugins.application_plugin import ApplicationPlugin

from poetry_multiverse_plugin.commands.check import CheckCommand
from poetry_multiverse_plugin.commands.info import InfoCommand
from poetry_multiverse_plugin.commands.lock import LockCommand
from poetry_multiverse_plugin.commands.run import RunCommand
from poetry_multiverse_plugin.commands.show import ShowCommand


def register(application: Application, *commands: Type[Command]):
    for cmd in commands:
        assert cmd.name is not None
        application.command_loader.register_factory(cmd.name, cmd)


class MultiversePlugin(ApplicationPlugin):
    def activate(self, application: Application):
        register(application, CheckCommand, InfoCommand, LockCommand, RunCommand, ShowCommand)



from concurrent.futures import ThreadPoolExecutor, as_completed
from cleo.helpers import option
from cleo.io.null_io import NullIO
from poetry.poetry import Poetry

from poetry_multiverse_plugin.commands.workspace import WorkspaceCommand
from poetry_multiverse_plugin.installers import Installers
from poetry_multiverse_plugin.workspace import Workspace


class LockCommand(WorkspaceCommand):
    name = 'workspace lock'
    description = 'Compute dependencies for the multiverse workspace.'

    options = [
        option(
            "no-update", None, "Reuse packages in project lock files."
        )
    ]

    help = """
The <info>workspace lock</info> command updates the <comment>poetry.lock</>
file for each project in the workspace such that dependencies required by
multiple projects are locked to the same version.

<info>poetry workspace lock</info>
"""

    def handle_workspace(self, workspace: Workspace) -> int:
        installer = Installers(workspace, NullIO(), self.env)
        return_code = 0

        with self.cli.progress('Locking workspace...'):
            return_code = installer.root(locked=self.option('no-update', True)).lock().run()
            if return_code != 0:
                self.io.write_error_line('<error>Failed to lock workspace!</>')
                return return_code

        with self.cli.status(None, 'Updating') as status:
            def run(project: Poetry) -> int:
                status(project).update('Locking...')
                return installer.project(project).lock().run()

            with ThreadPoolExecutor() as executor:
                jobs = {
                    executor.submit(lambda: run(project)): project
                    for project in workspace.projects
                }
                for future in as_completed(jobs):
                    project = jobs[future]
                    result = status(project).complete(future)
                    return_code = max(return_code, result if result is not None else 1)
        
        return return_code



from collections import defaultdict
from dataclasses import dataclass
from typing import Generic, Iterator, TypeVar
from poetry.core.packages.dependency import Dependency
from poetry.poetry import Poetry


RecordT = TypeVar('RecordT')


@dataclass
class Record(Generic[RecordT]):
    poetry: Poetry
    value: RecordT


@dataclass
class Dependencies:
    records: dict[str, list[Record[Dependency]]]

    @staticmethod
    def from_projects(*projects: Poetry) -> 'Dependencies':
        deps: dict[str, list[Record[Dependency]]] = defaultdict(list)
        for project in projects:
            for dep in project.package.all_requires:
                deps[dep.complete_name].append(Record(project, dep))
        return Dependencies(deps)

    def __iter__(self) -> Iterator[Dependency]:
        for package in self.records.values():
            constraint = package[0].value.constraint
            for dep in package[1:]:
                constraint = constraint.intersect(dep.value.constraint)
            yield package[0].value.with_constraint(constraint)
    
    @property
    def conflicts(self) -> Iterator[Dependency]:
        for dependency in self:
            if dependency.constraint.is_empty():
                yield dependency



from contextlib import _GeneratorContextManager, contextmanager
from dataclasses import dataclass
from queue import Queue
from threading import Thread
from typing import Callable, Iterator, Optional
from cleo.io.io import IO
from poetry.console.commands.installer_command import InstallerCommand

from poetry_multiverse_plugin.cli.progress import progress
from poetry_multiverse_plugin.cli.status import OutputQueue, StatusConfig, WorkspaceStatus
from poetry_multiverse_plugin.workspace import Workspace


@dataclass
class CliUtils:
    workspace: Workspace
    io: IO

    def progress(self, message: str) -> _GeneratorContextManager[None]:
        return progress(self.io, message)
    
    @contextmanager
    def status(self, header: Optional[str], action: str) -> Iterator[WorkspaceStatus]:
        q: Queue[Optional[Callable[[], None]]] = Queue()

        def process():
            while task := q.get():
                task()

        io_thread = Thread(target=process)
        try:
            io_thread.start()
            yield WorkspaceStatus(
                OutputQueue(q.put, self.io),
                self.workspace.projects,
                StatusConfig(header, StatusConfig.default_template(action))
            )
        finally:
            q.put(None)
            io_thread.join()


class WorkspaceCommand(InstallerCommand):
    @property
    def cli(self) -> CliUtils:
        workspace = Workspace.create(self.poetry)
        assert workspace is not None
        return CliUtils(workspace, self.io)

    def handle_workspace(self, workspace: Workspace) -> int:
        raise NotImplementedError
    
    def handle(self) -> int:
        workspace = Workspace.create(self.poetry)
        if not workspace:
            self.line_error('Unable to locate workspace root')
            return 1
        return self.handle_workspace(workspace)



from cleo.io.outputs.output import Output
from cleo.io.outputs.section_output import SectionOutput


def can_overwrite(output: Output) -> bool:
    return output.is_decorated() and not output.is_debug()


def overwrite(output: Output, message: str):
    if isinstance(output, SectionOutput):
        output.overwrite(message)
    else:
        output.write_line(message.strip())



import asyncio
from asyncio import StreamReader, create_subprocess_exec, subprocess
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from typing import List, Optional
from cleo.helpers import argument, option
from poetry.poetry import Poetry

from poetry_multiverse_plugin.commands.workspace import WorkspaceCommand
from poetry_multiverse_plugin.workspace import Workspace


class RunCommand(WorkspaceCommand):
    name = 'workspace run'
    description = 'Execute a command in every project in the workspace.'

    arguments = [
        argument(
            'program',
            description="The command to run in each project's environment",
            multiple=True
        )
    ]

    options = [
        option(
            'parallel', 'p',
            description='The maximum number of processes to run in parallel',
            flag=False
        )
    ]

    def handle_workspace(self, workspace: Workspace) -> int:
        command: List[str] = self.argument('program')
        parallel = self.option('parallel')

        with self.cli.status(f'\r\nRunning {" ".join(command)}', 'Running') as status:
            async def run(project: Poetry) -> int:
                status(project).update('Running...')

                async def _read_stream(stream: Optional[StreamReader], *, error: bool = False):  
                    if stream:
                        while line := await stream.readline():
                            status(project).log(line.decode('utf-8'), error=error)

                proc = await create_subprocess_exec(
                    *command,
                    cwd=project.pyproject_path.parent,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE
                )
                await asyncio.gather(
                    _read_stream(proc.stdout),
                    _read_stream(proc.stderr, error=True)
                )
                return await proc.wait()

            with ThreadPoolExecutor(int(parallel) if parallel else None) as executor:
                jobs = {
                    executor.submit(partial(lambda p: asyncio.run(run(p)), project)): project
                    for project in workspace.projects
                }

                return_code = 0

                for future in as_completed(jobs):
                    project = jobs[future]
                    result = status(project).complete(future)
                    return_code = max(return_code, result if result is not None else 1)
            
                return return_code



from concurrent.futures import Future
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, Iterable, Optional, TypeVar
from cleo.io.io import IO
from cleo.io.outputs.output import Output
from poetry.poetry import Poetry

from poetry_multiverse_plugin.cli.utils import can_overwrite, overwrite


ResultT = TypeVar('ResultT')


@dataclass
class OutputQueue:
    submit: Callable[[Callable[[], None]], None]
    io: IO

    def section(self) -> Output:
        if can_overwrite(self.io.output):
            return self.io.section()
        return self.io.output


@dataclass
class StatusConfig:
    header: Optional[str]
    template: str

    @staticmethod
    def default_template(action: str) -> str:
        return ' '.join([
            '  <fg={color};options=bold>-</>',
            action,
            '<fg=cyan>{project}</>',
            '(<fg={color}>{status}</>)'
        ])


@dataclass
class StatusOutput:
    submit: Callable[[Callable[[], None]], None]
    summary_section: Output
    log_section: Output

    def summary(self, message: str):
        self.submit(lambda: overwrite(self.summary_section, message))

    def log(self, message: str):
        self.submit(lambda: self.log_section.write_line(message.strip()))


@dataclass
class ProjectStatus:
    writer: StatusOutput
    template: str
    name: str

    def log(self, message: str, *, error: bool = False):
        fmt = '<error>' if error else '<debug>'
        message = f'{fmt}[{self.name}]</> {message.rstrip()}'
        self.writer.log(message)

    def update(self, status: str, *, color: str = 'blue'):
        message = self.template.format(status=status, color=color, project=self.name)
        self.writer.summary(message)
    
    def complete(self, result: Future[int]) -> Optional[int]:
        assert result.done()
        if result.cancelled():
            self.update('Canceled', color='gray')
        elif result.exception():
            self.update('Error', color='red')
        else:
            code = result.result()
            if code != 0:
                self.update('Failed', color='red')
            else:
                self.update('Done', color='green')
            return code
        return None


class WorkspaceStatus:
    def __init__(self, output: OutputQueue, projects: Iterable[Poetry], config: StatusConfig) -> None:
        self.log_section = output.section()
        if config.header:
            self.header_section = output.section()
            overwrite(self.header_section, config.header)
        self.sections: Dict[Path, ProjectStatus] = {
            project.pyproject_path: ProjectStatus(
                StatusOutput(output.submit, output.section(), self.log_section),
                config.template,
                project.package.name
            )
            for project in sorted(projects, key=lambda p: p.package.name)
        }
        for action in self.sections.values():
            action.update('Pending', color='blue')
    
    def __call__(self, project: Poetry) -> ProjectStatus:
        return self.sections[project.pyproject_path]



from typing import List
from poetry.core.packages.dependency import Dependency
from poetry.core.packages.package import Package

from poetry_multiverse_plugin.commands.workspace import WorkspaceCommand
from poetry_multiverse_plugin.dependencies import Dependencies, Record
from poetry_multiverse_plugin.packages import Packages
from poetry_multiverse_plugin.workspace import Workspace


class CheckCommand(WorkspaceCommand):
    name = 'workspace check'
    description = 'Validate dependencies and locked packages in the multiverse workspace.'

    def render_dependency_conflicts(self, dependencies: Dependencies, conflicts: List[Dependency]):
        def format_row(dep: Record[Dependency]) -> List[str]:
            return [
                f'  {dep.poetry.package.name}',
                f'<b>{dep.value.pretty_constraint}</>'
            ]
        
        for conflict in conflicts:
            records = sorted(
                dependencies.records[conflict.complete_name],
                key=lambda rec: rec.poetry.package.name
            )
            self.table(
                style='compact',
                header=conflict.complete_pretty_name,
                rows=[format_row(rec) for rec in records]
            ).render()

    def render_duplicate_packages(self, packages: Packages, duplicates: List[str]):
        def format_row(dep: Record[Package]) -> List[str]:
            return [
                f'  {dep.poetry.package.name}',
                f'<b>{dep.value.pretty_version}</>'
            ]
        
        for package_name in duplicates:
            package = packages.records[package_name][0].value
            records = sorted(
                packages.records[package_name],
                key=lambda rec: rec.poetry.package.name
            )
            self.table(
                style='compact',
                header=package.complete_pretty_name,
                rows=[format_row(rec) for rec in records]
            ).render()

    def handle_workspace(self, workspace: Workspace) -> int:
        dependencies = workspace.dependencies
        conflicts = sorted(
            dependencies.conflicts,
            key=lambda dep: dep.complete_pretty_name
        )

        if conflicts:
            self.render_dependency_conflicts(dependencies, conflicts)
            self.io.write_error_line('<error>Dependency constraint conflicts found!</>')
            return 1
        
        packages = workspace.packages
        duplicates = sorted(
            package
            for package, versions in packages.records.items()
            if len(set(record.value.version for record in versions)) > 1
        )

        if duplicates:
            self.render_duplicate_packages(packages, duplicates)
            self.io.write_error_line('<error>Multiple package versions found!</>')
            return 1

        self.io.write_line('<fg=green>No errors found!</>')
        return 0



from typing import Any, List, Optional, TypedDict
from poetry.poetry import Poetry

def resolve(doc: Any, *path: str) -> Optional[Any]:
    if not path:
        return doc
    if not isinstance(doc, dict):
        return None
    return resolve(doc.get(path[0]), *path[1:])


class WorkspaceConfig(TypedDict, total=False):
    root: bool
    projects: List[str]


def parse_config(poetry: Poetry) -> WorkspaceConfig:
    return WorkspaceConfig(
        resolve(poetry.pyproject.data.unwrap(), 'tool', 'multiverse') or {}
    )



from cleo.io.null_io import NullIO
from poetry.console.commands import show

from poetry_multiverse_plugin.installers import Installers
from poetry_multiverse_plugin.workspace import Workspace


class ShowCommand(show.ShowCommand):
    name = 'workspace show'
    description = 'List dependencies in the multiverse workspace.'

    def handle(self) -> int:
        workspace = Workspace.create(self.poetry)
        if not workspace:
            self.io.write_error('Unable to locate workspace root')
            return 1

        installer = Installers(workspace, NullIO(), self.env)
        installer.root(locked=True).lock().run()
        self.set_poetry(workspace.root)
        for dep in workspace.dependencies:
            self.poetry.package.add_dependency(dep)
        self.poetry.set_locker(installer.locker)
        return super().handle()



from typing import List
from poetry.poetry import Poetry

from poetry_multiverse_plugin.commands.workspace import WorkspaceCommand
from poetry_multiverse_plugin.workspace import Workspace


class InfoCommand(WorkspaceCommand):
    name = 'workspace info'
    description = 'Describe the multiverse workspace.'

    def handle_workspace(self, workspace: Workspace) -> int:
        root = workspace.root
        projects = sorted(workspace.projects, key=lambda p: p.package.name)

        def format_row(project: Poetry) -> List[str]:
            return [
                f'<b>{project.package.name}</>',
                str(project.pyproject_path.parent.relative_to(root.pyproject_path.parent))
            ]

        self.table(style='compact') \
            .set_headers([f'{root.package.name}', str(root.pyproject_path.parent)]) \
            .set_rows([format_row(project) for project in projects]) \
            .render()
        return 0



from pathlib import Path
from tempfile import _TemporaryFileWrapper, NamedTemporaryFile
from typing import Optional
from cleo.io.io import IO
from dataclasses import dataclass, field
from poetry.core.packages.project_package import ProjectPackage
from poetry.installation.installer import Installer
from poetry.packages.locker import Locker
from poetry.poetry import Poetry
from poetry.repositories.repository_pool import RepositoryPool, Priority
from poetry.utils.env import Env

from poetry_multiverse_plugin.workspace import Workspace


@dataclass
class Installers:
    workspace: Workspace
    io: IO
    env: Env
    lock_file: _TemporaryFileWrapper = field(default_factory=lambda: NamedTemporaryFile(suffix='.lock'))

    @property
    def locker(self) -> Locker:
        return Locker(Path(self.lock_file.name), {})

    def root(self, *, locked: bool) -> Installer:
        root = self.workspace.root
        aggregate_project = ProjectPackage(root.package.name, root.package.version)
        aggregate_project.python_versions = root.package.python_versions

        for dep in self.workspace.dependencies:
            aggregate_project.add_dependency(dep)

        return Installer(
            self.io,
            self.env,
            aggregate_project,
            self.locker,
            self.workspace.packages.repository_pool if locked else root.pool,
            root.config,
            disable_cache=root.disable_cache
        )
    
    def repository_pool(self, poetry: Optional[Poetry] = None) -> RepositoryPool:
        pool = RepositoryPool([self.locker.locked_repository()])

        if poetry:
            for repo in poetry.pool.all_repositories:
                pool.add_repository(repo, priority=Priority.SUPPLEMENTAL)

        return pool
    
    def project(self, poetry: Poetry) -> Installer:
        root = self.workspace.root

        return Installer(
            self.io,
            self.env,
            poetry.package,
            poetry.locker,
            self.repository_pool(poetry),
            poetry.config,
            disable_cache=root.disable_cache
        )



from dataclasses import dataclass
import os
from pathlib import Path
from poetry.factory import Factory
from poetry.poetry import Poetry
from typing import Iterable, Optional

from poetry_multiverse_plugin.config import WorkspaceConfig, parse_config
from poetry_multiverse_plugin.dependencies import Dependencies
from poetry_multiverse_plugin.packages import Packages


def workspace_root(poetry: Poetry) -> Optional[Poetry]:
    if parse_config(poetry).get('root') is True:
        return poetry
    try:
        parent_project = Factory().create_poetry(
            poetry.pyproject_path.resolve().parent.parent,
            disable_cache=poetry.disable_cache
        )
        return workspace_root(parent_project)
    except RuntimeError:
        return None


@dataclass
class Workspace:
    root: Poetry

    @staticmethod
    def create(poetry: Poetry) -> Optional['Workspace']:
        root = workspace_root(poetry)
        return Workspace(root) if root else None
    
    @property
    def config(self) -> WorkspaceConfig:
        return parse_config(self.root)
    
    @property
    def project_dirs(self) -> Iterable[Path]:
        project_dirs = self.config.get('projects', ['**'])
        for dir_glob in project_dirs:
            toml_glob = f'{dir_glob.rstrip(os.sep)}{os.sep}pyproject.toml'
            for pyproject in self.root.pyproject_path.parent.glob(toml_glob):
                yield pyproject.parent.resolve(True)
    
    @property
    def projects(self) -> Iterable[Poetry]:
        for project in self.project_dirs:
            if project == self.root.pyproject_path.parent:
                continue
            try:
                yield Factory().create_poetry(
                    project,
                    disable_cache=self.root.disable_cache
                )
            except RuntimeError:
                print(f'Skipping pyproject.toml {project}')
    
    @property
    def dependencies(self) -> Dependencies:
        return Dependencies.from_projects(self.root, *self.projects)
    
    @property
    def packages(self) -> Packages:
        return Packages.from_projects(self.root, *self.projects)
